/**
  Aurora Rendering demo application.
  
  Orion Sky Lawlor, olawlor@acm.org, 2010-09-22 (Public Domain)
*/
#include <iostream>
#include <cmath>
#include <GL/glew.h> /* 
OpenGL Extensions Wrangler:  for gl...ARB extentions.  Must call glewInit after glutCreateWindow! */
#include <GL/glut.h> /* OpenGL Utilities Toolkit, for GUI tools */
#include "ogl/glsl.h"
#include "ogl/glsl.cpp"
#define USE_OGL_JOYSTICK 1 /* joystick makes for very smooth camera motion */
#include "ogl/minicam.h"
#include "osl/mat4.h"
#include "osl/mat4_inverse.h"
#include "ogl/dumpscreen.h"
#include "ogl/pixelbench.h"

const float km=1.0/6371.0; // convert kilometers to render units (planet radii)
int benchmode=0;

/** SOIL **/
#include "soil/SOIL.h" /* Simple OpenGL Image Library, www.lonesock.net/soil.html (plus Dr. Lawlor
's modifications) */
#include "soil/SOIL.c" /* just slap in implementation files here, for easier linking */
#include "soil/stb_image_aug.c"
void read_soil_jpeg(const char *filename, GLenum mode=GL_RGBA8)
{
	printf("Reading texture '%s'",filename); fflush(stdout);
	GLuint srcTex = SOIL_load_OGL_texture
	(
		filename,
		SOIL_LOAD_AUTO,
		SOIL_CREATE_NEW_ID,
		SOIL_FLAG_MIPMAPS | SOIL_FLAG_INVERT_Y,
		mode
	);
	printf(".\n");
	glBindTexture(GL_TEXTURE_2D,srcTex);
}

#include <vector>

// A quadric object
struct quadric_object {
	osl::mat4 A; // quadric surface matrix: 0 = x^T A x
	
	vec3 color; // diffuse color
	float reflect; // 0=matte; 1=totally specular reflective
	
	osl::mat4 MVinv; // from world to trim space
	
	float refract; // 0=matte; 1=totally transparent
	float refract_index; // refractive index (if refractive)
	float checkerboard; // 0=uniform color; 1=black and white checkers
	float slice; // slices per meter
	
	//float opacity; // 0=transparent; 1=opaque
};

void upload_modelview(osl::mat4 &A,quadric_object &q) 
{
	// Object matrix
	osl::mat4 mv; // object -> world space
	glGetFloatv(GL_MODELVIEW_MATRIX,&mv[0][0]);
	osl::mat4 B=inverse(mv); // world -> object
	q.MVinv=B;
	osl::mat4 BAB=transpose(B)*A*B; // see rule to modify quadric object
	q.A=BAB; // upload quadric matrix
}

void display(void) 
{
	glEnable(GL_DEPTH_TEST);
	glDepthMask(GL_TRUE); // turn on Z writes
	glEnable(GL_BLEND);
	
	glClearColor(0.4,0.5,0.7,0.0); // background color
	glClear(GL_COLOR_BUFFER_BIT + GL_DEPTH_BUFFER_BIT);
	
	glViewport(0,0, glutGet(GLUT_WINDOW_WIDTH),glutGet(GLUT_WINDOW_HEIGHT));
	
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity(); // flush any ancient matrices
	gluPerspective(70.0, // fov
		glutGet(GLUT_WINDOW_WIDTH)/(float)glutGet(GLUT_WINDOW_HEIGHT),
		0.1,
		1000.0); // z clipping planes
	float altitude=length(camera);
	float cameraVelocity=0.5*(altitude-0.99);
	oglCameraLookAt(cameraVelocity);
	float min_alt=1.0+1.0e-4; /* minimum allowed altitude */
	if (length(camera)<min_alt) camera=normalize(camera)*min_alt; // boot out of planet
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity(); // flush any ancient matrices
	
	/* The proxy sphere should be gently rotating */
	const float rotate_rate=0.01; /* revolutions per second */
	double time_in_seconds=glutGet(GLUT_ELAPSED_TIME)*0.001;
	float angle=time_in_seconds*rotate_rate*360;
	
	/* Build the pixel shader */
	static GLhandleARB prog=makeProgramObjectFromFiles(
		"raytrace_vertex.txt","raytrace.txt");
	glUseProgramObjectARB(prog);
	glFastUniform3fv(prog,"C",1,camera);
	
	static bool read_imgs=true; /* only read textures on the first frame */
/* Upload planet texture, to texture unit 1 */
	glActiveTexture(GL_TEXTURE1);
	if (read_imgs) read_soil_jpeg("tex/nightearth.jpg",GL_LUMINANCE8);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAX_ANISOTROPY_EXT,4);
	//glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
	glFastUniform1i(prog,"nightearthtex",1); // texture unit number
	
/* Upload aurora source texture, to texture unit 2 */
	glActiveTexture(GL_TEXTURE2);
	if (read_imgs) read_soil_jpeg("tex/aurora.jpg",GL_RGBA8);
	//glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAX_ANISOTROPY_EXT,4);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP_TO_EDGE);
	glFastUniform1i(prog,"auroratex",2); // texture unit number
	
/* Upload aurora distance texture, to texture unit 3 */
	glActiveTexture(GL_TEXTURE3);
	if (read_imgs) read_soil_jpeg("tex/aurora_distance.jpg",GL_LUMINANCE8);
	/* mipmaps screw up distance estimates at bottom of curtains. */
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP_TO_EDGE);
	glFastUniform1i(prog,"auroradistance",3); // texture unit number
	
/* Upload deposition lookup texture, to texture unit 4 */
	glActiveTexture(GL_TEXTURE4);
	if (read_imgs) read_soil_jpeg("tex/deposition.bmp",GL_RGBA8);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP_TO_EDGE);
	glFastUniform1i(prog,"auroradeposition",4); // texture unit number
	
	read_imgs=false;
	glActiveTexture(GL_TEXTURE0);
	
	glFastUniform1f(prog,"time_uniform",0.001*glutGet(GLUT_ELAPSED_TIME));
	
	// Only draw one set of faces, to avoid drawing raytraced geometry twice.
	//   (front side may get clipped if you're inside the object, so draw back)
	glEnable(GL_CULL_FACE); glCullFace(GL_FRONT);
	glBlendFunc(GL_ONE,GL_ONE_MINUS_SRC_ALPHA); // premultiplied alpha
	
	// Draw one *HUGE* sphere; this one proxy geometry covers all our objects.
	//   Keep in mind the far clipping plane in determining the sphere radius
	glutSolidSphere(200.0,3,3); 
	
	glUseProgramObjectARB(0);
	
	if (benchmode==2) {oglPixelBench(display,64,4); return;}
	
	static int movie_mode=0;
	static double last_movieframe=0.0;
	static double frame_interval=1.0/24; /* time between movie frames, seconds */
	if (key_down['p'] || movie_mode || key_down['m']) {
		if (movie_mode) {
			if (last_movieframe+frame_interval<time_in_seconds) {
				oglDumpScreen();
				last_movieframe+=frame_interval;
			}
		} else {
			oglDumpScreen();
			last_movieframe=time_in_seconds;
		}
		key_down['p']=false; /* p is momentary print-screen; m is for movie */
		if (key_down['m']) movie_mode=1;
		if (key_down['p']) movie_mode=0;
	} 
	glutPostRedisplay(); // continual animation
	
	glutSwapBuffers();

/* Estimate framerate */
	static double last_time=0.0;
	static int framecount=0, last_framecount=0;
	framecount++;
	double cur_time=0.001*glutGet(GLUT_ELAPSED_TIME);
	if (cur_time>0.5+last_time) 
	{ // every half a second, estimate fps
		double time=cur_time-last_time;
		double time_per_frame=time/(framecount-last_framecount);
		if (benchmode==1) {
			static FILE *f=fopen("bench.txt","a");
			static int bcount=0;
			int w=glutGet(GLUT_WINDOW_WIDTH),h=glutGet(GLUT_WINDOW_HEIGHT);
			/* count	resolution	frames/sec	milliseconds/frame        nanoseconds/pixel */
			fprintf(f,"%d	%dx%d		%.2f	%.5f	%.3f\n",
				bcount,w,h, 1.0/time_per_frame,
				time_per_frame*1.0e3, time_per_frame/(w*h)*1.0e9);
			if (bcount++>3) {
				fclose(f);
				int err=system("cat bench.txt");
				oglDumpScreen();
				exit(err);
			}
		}
		else { /* not a benchmark, just an ordinary run */
			char str[100];
			sprintf(str,"Planetary Renderer: %.1f fps, %.1f ms/frame (%.1f km)",
				1.0/time_per_frame,1.0e3*time_per_frame,
				(altitude-1.0)/km);
#ifndef MPIGLUT_H
			printf("%s\n",str);

			glutSetWindowTitle(str);

			printf("Minicam: camera=vec3(%f,%f,%f); camera_orient.z=vec3(%f,%f,%f);\n",
				VEC3_TO_XYZ(camera),
				VEC3_TO_XYZ(camera_orient.z));
#endif
		}
		last_framecount=framecount;
		last_time=cur_time;
	}
}

int main(int argc,char *argv[]) 
{
	glutInit(&argc,argv);
	
	// This is high definition 720p
	int w=1280, h=720;
	//int w=1280/4, h=720/4;
#ifdef MPIGLUT_H
	w=8000; h=4000;
#endif
	for (int argi=1;argi<argc;argi++) {
		if (0==strcmp(argv[argi],"-bench")) benchmode=1;
		else if (0==strcmp(argv[argi],"-pixelbench")) benchmode=2;
		else if (2==sscanf(argv[argi],"%dx%d",&w,&h)) {}
		else printf("Unrecognized argument '%s'!\n",argv[argi]);
	}
	
	glutInitDisplayMode(GLUT_RGBA + GLUT_DEPTH + GLUT_DOUBLE);
	glutInitWindowSize(w,h);
	glutCreateWindow("Raytraced Planet");
	
	glewInit();
	
	glutDisplayFunc(display);
	oglCameraInit();
	//camera=vec3(0,-1.5,0.6); // over continental US
	camera=vec3(-0.3,-0.8,1.1); // over Alaska
	
	// Benchmark: v31, 60km altitude, over Finland, looking toward Iceland:
	camera=vec3(-0.467287,-0.223999,0.868115); 
	camera_orient.z=vec3(0.177043,-0.919557,-0.350815);
	camera_orient.y=camera; /* point up from planet */
	
	correct_head_tilt=false; // allow free flight...
	correct_head_tilt_sphere=true; // ...around sphere
	
	glutMainLoop();
	return 0;
}
